
#ifndef interrupt_h
#define interrupt_h

#include "header\GBA.h"
#include "header\IRQ.h"

#define IRQIndex_VBlank 0
#define IRQIndex_HBlank 1
#define IRQIndex_VCount 2
#define IRQIndex_Timer0 3
#define IRQIndex_Timer1 4
#define IRQIndex_Timer2 5
#define IRQIndex_Timer3 6
#define IRQIndex_DMA0 7
#define IRQIndex_DMA1 8
#define IRQIndex_DMA2 9
#define IRQIndex_DMA3 10
#define IRQIndex_Keyboard 11

static void IRQ_Setup(u16 _REG_IE,u16 _REG_DISPSTAT);

__attribute__ ((naked)) // devkitadv̊JnIXLbv
__attribute__ ((section (".data.iwram"))) void IRQ_Dummy(void)
{
  asm("bx lr");
}

u32 IRQTable[]=
{
  (u32)IRQ_Dummy, // VBlank
  (u32)IRQ_Dummy, // HBlank
  (u32)IRQ_Dummy, // VCount
  (u32)IRQ_Dummy, // Timer0
  (u32)IRQ_Dummy, // Timer1
  (u32)IRQ_Dummy, // Timer2
  (u32)IRQ_Dummy, // Timer3
  (u32)IRQ_Dummy, // DMA0
  (u32)IRQ_Dummy, // DMA1
  (u32)IRQ_Dummy, // DMA2
  (u32)IRQ_Dummy, // DMA3
  (u32)IRQ_Dummy // Keyboard
};

__attribute__ ((naked)) // devkitadv̊JnIXLbv
__attribute__ ((section (".data.iwram"))) void IRQ_Handler()
{
  // gėǂWX^ R0-R3,R12(IP)
  
  // JninƑ΁j
  asm("mov r2, #0x4000000"); // r2=REG_BASE
  asm("ldr r3, [r2,#0x200]!"); // r3IEIF r2+=0x200;
  asm("ldrh r1, [r2, #0x8]"); // r1IME r1=*(r2+0x8);
  asm("mrs r0,SPSR"); // r0=SPSR;
  asm("stmfd sp!,{r0-r2,lr}"); // push {SPSR,IME,0x4000200,lr}
  
  // REG_IME1BiȂł킩ȂǁB0ɂȂĂ\Hj
  asm("mov r0, #0x01"); // r0=0x01;
  asm("strh r0, [r2, #0x8]"); // *(r2+0x8)=r0;
  
  // REG_IEREG_IFANDBFh~H
  asm("and r1, r3, r3, lsr #16"); // r1=r3&(r3>>16);
  
  // 荞ݗv
  asm("mov r12, #0x04000000");
  asm("ldr r12, [r12,#-0x08]"); // r12=*0x03007ff8; = &IntrTable[0];
  asm("ands r0, r1, #1"); // V-blank interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x2"); // H-blank interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x4"); // V-counter interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x8"); // Timer 0 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x10"); // Timer1 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x20"); // Timer2 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x40"); // Timer3 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x100"); // DMA0 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x200"); // DMA1 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x400"); // DMA2 interrupt
  asm("bne jump_intr");
  
  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x800"); // DMA3 interrupt
  asm("bne jump_intr");

  asm("add r12,r12, #4");
  asm("ands r0, r1, #0x1000"); // Keys interrupt
  asm("bne jump_intr");
  
  // `荞݂疳[v
  asm("loop: b loop");
  
  // Jn
  asm("jump_intr:");
  
  asm("strh r0, [r2, #2]"); // REG_IFr0i荞ݗvrbgjZbgăNA

  // CPUʏ탂[hɂIRQLɂB
  asm("mrs r0,CPSR");
  asm("bic r0,r0,#0xdf");
//  asm("orr r0,r0,#0x92"); // SingleInt IRQMode
  asm("orr r0,r0,#0x1f"); // MultiInt UserMode
  asm("msr CPSR,r0");
  
  asm("ldr r0, [r12]"); // r12iIntrTable[IRQType]jr0ɓǂݍ
  
  // ֐s
  asm("stmfd sp!, {lr}");
  asm("adr lr, IntrRet");
  asm("bx r0");
  asm("IntrRet:");
  asm("ldmfd sp!, {lr}");
  
  // CPUIRQ[hɖ߂IRQ𖳌ɂB
  asm("mrs r0,CPSR");
  asm("bic r0,r0,#0xdf");
  asm("orr r0,r0,#0x92");
  asm("msr CPSR,r0");
  
  // niJnƑ΁j
  asm("ldmfd sp!, {r0-r2,lr}"); // pop {SPSR,IME,REG_BASE+0x200,lr}
  asm("strh r1, [r2, #0x8]"); // REG_IME𕜋A
  asm("msr spsr, r0"); // SPSR𕜋A
  
  // AB
  asm("bx lr");
}

void IRQ_Setup(u16 _REG_IE,u16 _REG_DISPSTAT)
{
  REG_IME = IRQ_MASTER_OFF;

  REG_INTERUPT = (u32)IRQ_Handler;
  
  // IWRAM̈e[uQƃ|C^ƂĎgBBIOSŎgȂ̈H
  u32 *pIRQTable=(u32*)0x03007ff8;
  *pIRQTable=(u32)&IRQTable[0];
  
  REG_IE=_REG_IE;
  REG_IF=0x3fff;
  REG_DISPSTAT=_REG_DISPSTAT;
  
  REG_IME=IRQ_MASTER_ON;
}

#endif
